#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/if.h>
#include <linux/inetdevice.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include "ipt_MNAT.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jens-Wolfhard Schicke, Bjoern Willenberg");
MODULE_DESCRIPTION("SNAT based upon uid");

#if 1
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define ERREXIT(ret, format, args...) do {\
	DEBUGP(format, args); \
	return ret; \
} while(0)

static int
mnat_check(const char *tablename,
	      const struct ipt_entry *e,
	      void *targinfo,
	      unsigned int targinfosize,
	      unsigned int hook_mask)
{
	struct ipt_mnat_template *tmpl = targinfo;

	if (strcmp(tablename, "nat") != 0) {
		ERREXIT(0, "MNAT: bad table `%s' - MNAT is only valid in nat table\n", tablename);
	} else if (targinfosize != IPT_ALIGN(sizeof(*tmpl))) {
		ERREXIT(0, "MNAT: size %u - internal error\n", targinfosize);
	} else if (hook_mask & ~(1 << NF_IP_POST_ROUTING)) {
		ERREXIT(0, "MNAT: bad hooks %x - internal error\n", hook_mask);
	}

	return 1;
}

static void 
mnat_destroy(void *targinfo,
		unsigned int targinfosize)
{
}

static unsigned int
mnat_target(struct sk_buff **pskb,
		const struct net_device *in,
		const struct net_device *out,
		unsigned int hooknum,
		const void *targinfo,
		void *userinfo)
{
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;
	u_int32_t new_ip;
	uint16_t owner;
	int pos;
	const struct ipt_mnat_template *tmpl = targinfo;
	const struct ip_conntrack_tuple *t;

	IP_NF_ASSERT(hooknum == NF_IP_POST_ROUTING);
	ct = ip_conntrack_get(*pskb, &ctinfo);

	t = &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;

	if(!pskb || !*pskb || !(*pskb)->sk || !(*pskb)->sk->sk_socket || !(*pskb)->sk->sk_socket->file)
		ERREXIT(0, "MNAT: unexpected NULL pointer: %d\n", 0);
	owner = (*pskb)->sk->sk_socket->file->f_uid;
	DEBUGP("owner: %X\n", owner);

	new_ip = 0;
	for(pos = 0; pos < 4; pos++) {
		uint16_t new_ipbyte;
		struct ipt_mnat_byte_template *postmpl = tmpl->byte_template + pos;
		new_ipbyte = owner;
		new_ipbyte /= postmpl->divisor;
		new_ipbyte %= postmpl->base;
		new_ipbyte += postmpl->offset;
		new_ip += (new_ipbyte & 0xFF) << (8 * pos);

		DEBUGP("template pos %d: div %d mod %d offset %d\n",
				pos,
				postmpl->divisor, 
				postmpl->base, 
				postmpl->offset);
	}
	new_ip = htonl(new_ip);

	DEBUGP("MNAT: src=%u.%u.%u.%u dst=%u.%u.%u.%u, "
			"new src=%u.%u.%u.%u\n",
			NIPQUAD(t->src.ip), NIPQUAD(t->dst.ip),
			NIPQUAD(new_ip));

	/* Transfer from original range. */
	struct ip_nat_multi_range newrange = ((struct ip_nat_multi_range)
		{ 1, { { IP_NAT_RANGE_MAP_IPS,
			 new_ip, new_ip,
			 { 0x0000 } , { 0xFFFF } } } });

	/* Hand modified range to generic setup. */
	return ip_nat_setup_info(ct, &newrange, hooknum);
}

static struct ipt_target mnat_reg = { 
	.name		= "MNAT",
	.target		= mnat_target,
	.checkentry	= mnat_check,
	.destroy	= mnat_destroy,
	.me		= THIS_MODULE,
};

static int __init init(void)
{
	return ipt_register_target(&mnat_reg);
}

static void __exit fini(void)
{
	ipt_unregister_target(&mnat_reg);
}

module_init(init);
module_exit(fini);
